#include <Crypto.h>
#include <AES.h>

#include <Arduino_CRC32.h>

#include <Adafruit_PN532.h>

#define IRQ 9

Adafruit_PN532 nfc(IRQ,100);

uint8_t protected_programm[13][4]; //  Защищенный код
uint8_t unprotected_programm[16][4]; // Незащищенный код

Arduino_CRC32 crc;
AESTiny128 aes128;

const uint8_t enc_key[] PROGMEM = {0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07,0x08, 0x09, 0x0A, 0x0B, 0x0C, 0x0D, 0x0E, 0x0F}; // Ключ шифрования
const uint8_t unprot_magic[] PROGMEM = {0xFF,0xAA,0xFF,0xAA}; // "Магическая" комбинация для определения наличия незащищенного кода

const uint8_t unprot PROGMEM = 0; // константа, если равна единице, программатор будет записывать в незащищенную секцию

void setup() {
  //Конфигурация оборудования
  Serial.begin(9600);
  Serial.println(F("Initialization start"));
  nfc.begin();
  uint32_t versiondata = nfc.getFirmwareVersion();
 if (!versiondata) {
    Serial.println(F("Didn't find PN53x board"));
    while (1); // halt
  }
  nfc.SAMConfig();

  // Запись программы в карту (пока вручную)
  protected_programm[0][0] = 0xAA;
  protected_programm[0][1] = 0x03;
  protected_programm[0][2] = 0xFF;
  protected_programm[0][3] = 0x00;
}

void loop() {

    //Ожидание ввода
    if(!Serial.available()){
      return;
    }else{
      while(Serial.available()){
        Serial.read();
      }
      Serial.println("Starting writing");
    }
    
    uint8_t success;
    uint8_t uidLength;
    uint8_t uid[] = { 0, 0, 0, 0, 0, 0, 0 }; 
    uint8_t zero[] = {0, 0, 0, 0};
    success = nfc.readPassiveTargetID(PN532_MIFARE_ISO14443A, uid, &uidLength); // Поиск карты
    if(success && uidLength == 7){ // Карта найдена, запись защищенного кода
      for(uint8_t i=0;i<13;i++){
        Serial.print("Writing ");
        Serial.print(i,DEC);
        Serial.println("/13 protected page");
        if(!nfc.ntag2xx_WritePage(4+i,protected_programm[i])){
          Serial.println("Write error");
        }
      }
      if(unprot){ // Запись незащищенного кода
        nfc.ntag2xx_WritePage(17,unprot_magic);
        for(uint8_t i=18;i<35;i++){
          Serial.print("Writing ");
          Serial.print(i-18,DEC);
          Serial.println("/16 unprotected page"); 
          if(!nfc.ntag2xx_WritePage(i,(uint8_t*)unprotected_programm[i-18])){
            Serial.println("Write error");
          }
        }
        // Вычисление и шифрование проверочной суммы
        uint32_t checksum = crc.calc(&unprotected_programm[0][0],16*4*sizeof(uint8_t)); 
        uint32_t pad = 0xDEADBEEF; // Значение для заполнения оставшегося буфера
        uint32_t enc_buffer[] = {checksum,pad,pad,pad};
        uint8_t encrypted[16]; 
        crypto_feed_watchdog();
        aes128.setKey(enc_key,aes128.keySize());
        aes128.encryptBlock(encrypted,(uint8_t*)&enc_buffer[0]);
        //Запись
        for(uint8_t i=0;i<4;i++){
             nfc.ntag2xx_WritePage(35+i,&encrypted[i*4]); // 
        }
        Serial.println("Written protection"); 
      }else{
        nfc.ntag2xx_WritePage(17,zero);
      }
    }
}
