#include <Arduino_CRC32.h>

//"Ядро" эмулятора, для подтверждения целостности незащищенных страниц используется следующий алгоритм:
//1. Из незащищенного кода получается crc сумма
//2. Сумма шифруется алгоритмом AES128
//3. Полученное значение сравнивается с эталонным, записанным на последней странице памяти
//4. Если значения не совпали выполняется защищенный код

//Для экономии памяти использовал побитовое сжатие ( |=; &= ~ и др.), что это объясню завтра

//Protected code - минимальная программа для защищенной памяти - гарантирует что машина будет стоять на месте
//0xAA 0x00 0x00 0xF0
//0x00 0xFF 0x00 0x00

//Example 1 - просто движение вперед до останавливающего знака
//0xAA 0x01 0xFF 0x00

//Example 2 - поворот на 120 градусов затем движение вперед до останавливающего знака
//0xAB 0x78 0xAA 0x03
//0xAA 0x01 0x00 0xFF

//Stop_Sign - останавливающий знак
//0xAA 0x00 0x00 0xFF 

#include <Crypto.h>
#include <AES.h>

#include <Adafruit_PN532.h>



#define OPFLAG_NOARG 1 

#define EXECFLAG_PROGSET 1
#define EXECFLAG_CMP     2

#define SPEED_1      5 
#define DIR_1        4
 
#define SPEED_2      6
#define DIR_2        7

#define IRQ 9

Adafruit_PN532 nfc(IRQ, 100);

typedef struct{
  uint8_t flags;
  uint8_t op;
  uint8_t (*handler)(uint8_t arg); 
}op_t;

op_t* ops[32];
uint32_t op_c;

uint8_t reg_exec_flags;

uint8_t reg_dat[16];

uint8_t reg_sel;
uint8_t reg_cmp;
uint8_t reg_ang;

uint32_t reg_slp;
uint32_t reg_eip;

uint8_t c_programm[16][4];
op_t* reg_cop;
uint8_t reg_carg;

Arduino_CRC32 crc;
AESTiny128 aes128;

const uint8_t enc_key[] PROGMEM = {0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07,0x08, 0x09, 0x0A, 0x0B, 0x0C, 0x0D, 0x0E, 0x0F};
const uint8_t unprot_magic[] PROGMEM = {0xFF,0xAA,0xFF,0xAA};

void clean_registers(){
  
  reg_eip = 0;
  
  memset(reg_dat,0,16*sizeof(uint8_t));

  reg_cop =  0;
  reg_carg = 0;

  reg_sel = 0;
  reg_cmp = 0;
  reg_ang = 0;

  for(uint8_t i = 0;i < 16; i++){
    memset(&c_programm[i][0],0,4);
  }
}

uint8_t echo_handlr(uint8_t arg){
  Serial.println(arg,HEX);
  return 0;
}

uint8_t null_handlr(uint8_t arg){
  return 0;
}

uint8_t move_handlr(uint8_t arg){
  //arg = {1;0;2;3;4} - direction

  switch(arg){
    case 0: //STOP
     analogWrite(SPEED_1, 0);
     analogWrite(SPEED_2, 0);
     return 0;
    case 1://FORWARD
     digitalWrite(DIR_1, LOW);
     digitalWrite(DIR_2, LOW);
     analogWrite(SPEED_1, 210);
     analogWrite(SPEED_2, 140);
     return 0;
    case 2://BACKWARD
     digitalWrite(DIR_1, HIGH);
     digitalWrite(DIR_2, HIGH);
     analogWrite(SPEED_1, 210);
     analogWrite(SPEED_2, 140);
     return 0;
    case 3://LEFT
     digitalWrite(DIR_1, HIGH);
     digitalWrite(DIR_2, LOW);
     analogWrite(SPEED_1, 0);
     analogWrite(SPEED_2, 140);
     return 0;
    case 4://RIGHT
     digitalWrite(DIR_1, LOW);
     digitalWrite(DIR_2, HIGH);
     analogWrite(SPEED_1, 140);
     analogWrite(SPEED_2, 0);
     return 0;
    default:
     return 1;
  }
}

uint8_t sav_handlr(uint8_t arg){
  reg_dat[reg_sel] = arg;
  return 0;
}

uint8_t inc_handlr(uint8_t arg){
  reg_dat[reg_sel] += arg;
  return 0;
}

uint8_t dec_handlr(uint8_t arg){
  reg_dat[reg_sel] -= arg;
  return 0;
}

uint8_t sleep_handlr(uint8_t arg){
  reg_slp = arg;
  return 0;
}

uint8_t end_handlr(uint8_t arg){
  reg_exec_flags &= ~EXECFLAG_PROGSET; 
  return 0;
}

uint8_t jmp_handlr(uint8_t arg){
  if(arg >= 64){
    return 1;
  }
  reg_eip = arg;
  return 0;
}

uint8_t jmpif_handlr(uint8_t arg){
  if(reg_exec_flags & EXECFLAG_CMP){
    reg_exec_flags &= ~EXECFLAG_CMP;
    return jmp_handlr(arg);
  }
  return 0;
}

uint8_t cmp_handlr(uint8_t arg){
 switch(reg_cmp){
    case 0:
      if(reg_dat[arg] < reg_cmp){
        reg_exec_flags |= EXECFLAG_CMP;
      }
    break;
    case 1:
      if(reg_dat[arg] <= reg_cmp){
        reg_exec_flags |= EXECFLAG_CMP;
      }
    break;
    case 2:
      if(reg_dat[arg] > reg_cmp){
        reg_exec_flags |= EXECFLAG_CMP;
      }
    break;
    case 3:
      if(reg_dat[arg] >= reg_cmp){
        reg_exec_flags |= EXECFLAG_CMP;
      }
    break;
    case 4:
      if(reg_dat[arg] == reg_cmp){
        reg_exec_flags |= EXECFLAG_CMP;
      }
    break;
 }
 return 0;
}

uint8_t setcmp_handlr(uint8_t arg){
  reg_cmp = arg;
  return 0;
}

uint8_t sel_handlr(uint8_t arg){
  reg_sel = arg;
}

void register_op(uint8_t op,uint8_t (*handler)(uint8_t arg),uint8_t need_arg){
  //Serial.println(op,HEX);
  if(find_op(op)){
    Serial.print(F("Failed to register: Operator 0x"));
    Serial.print(op,HEX);
    Serial.println(F(" already registered!"));
    return;
  }
  if(op_c >= 32){
    Serial.println(F("Failed to register: Operator limit exceeded!"));
    return;
  }
  op_t* op_s = (op_t*)malloc(sizeof(op_t));
  op_s->op = op;
  op_s->handler = handler;
  op_s->flags = 0;
  if(!need_arg){
    op_s->flags |= OPFLAG_NOARG;
  }
  ops[op_c] = op_s;
  op_c++;
}

op_t* find_op(uint8_t op){
  for(uint32_t i=0;i<op_c;i++){
    if(ops[i]->op == op){
      return ops[i];
    }
  }
  return 0;
}

uint8_t process_op(op_t* op){
  Serial.print(F("Processing "));
  Serial.print(op->op,HEX);
  Serial.print("(");
  Serial.print(reg_carg);
  Serial.println(")");
  return op->handler(reg_carg);
}

void setup() {
  Serial.begin(9600);
  Serial.println(F("Initialization start"));
  nfc.begin();
  uint32_t versiondata = nfc.getFirmwareVersion();
 if (!versiondata) {
    Serial.println(F("Can't find PN53x board"));
    while (1); // halt
  }

  for (int i = 4; i < 8; i++) {     
    pinMode(i, OUTPUT);
  }
  
  nfc.SAMConfig();
  op_c = 0;

  clean_registers();

  memset(ops,0,32*sizeof(op_t*));

  // 480 bytes - 684 left (~28 operators)
  
  register_op(0x00,null_handlr ,0); // NOP
  register_op(0x0D,echo_handlr ,1); // Отладочный вывод
  register_op(0xAA,move_handlr ,1); // Движение
  register_op(0xA0,sav_handlr  ,1); // Сохранение значения в один их трех регистров общего назначения
  register_op(0xA1,inc_handlr  ,1); // Увеличение значения регистра
  register_op(0xA2,dec_handlr  ,1); // Уменьшение значения регистра
  register_op(0xCC,cmp_handlr  ,1); // Сравнение 
  register_op(0xCA,setcmp_handlr,1);// Операнд сравнения
  register_op(0xDD,sleep_handlr,1); // Задержка
  register_op(0xC0,sel_handlr,  1); // Выбор регистра
  register_op(0xFE,jmpif_handlr,1); // Условный переход
  register_op(0xF0,jmp_handlr  ,1); // Безусловный переход
  register_op(0xFF,end_handlr  ,0); // Выход из программы
  
  Serial.println(F("Initialization end"));
}

void loop() {
  if(!(reg_exec_flags & EXECFLAG_PROGSET)){
    uint8_t success;
    uint8_t uidLength;
    uint8_t uid[] = { 0, 0, 0, 0, 0, 0, 0 }; 
    success = nfc.readPassiveTargetID(PN532_MIFARE_ISO14443A, uid, &uidLength);
    if(success && uidLength == 7){
      
     
      clean_registers();

      uint8_t   protected_programm[13][4];
      uint8_t unprotected_programm[16][4];
      
      for(uint8_t i=0;i<13;i++){
        nfc.ntag2xx_ReadPage(4+i,&protected_programm[i][0]);
        
      }
      
      uint8_t magic[4];
      nfc.ntag2xx_ReadPage(17,magic); 
      uint8_t res = 1;
      for(uint8_t i=0;i<4;i++){
        if(magic[i] != unprot_magic[i]){
          res = 0;
          break;
        }
      }
      if(res){
        
        Serial.println(F("Found unprotected code"));
       // Serial.println(magic,HEX);
        for(uint8_t i=18;i<35;i++){
          nfc.ntag2xx_ReadPage(i,unprotected_programm[i-18]);
        }

        Serial.println(F("Verifying..."));
        uint32_t checksum = crc.calc(&unprotected_programm[0][0],16*4*sizeof(uint8_t));
        uint32_t pad = 0xDEADBEEF;
        uint32_t enc_buffer[] = {checksum,pad,pad,pad};
        uint8_t encrypted[16]; 
        crypto_feed_watchdog();
        aes128.setKey(enc_key,aes128.keySize());
        aes128.encryptBlock(encrypted,(uint8_t*)&enc_buffer[0]);
        uint8_t eta[16];
        for(uint8_t i=0;i<4;i++){
             nfc.ntag2xx_ReadPage(35+i,&eta[i*4]);
        }
        if(memcmp(eta,encrypted,16)){
          Serial.println(F("Failed to verify code integrity, executing protected code!"));
          for(int i=0;i<13;i++){
             memcpy(c_programm[i],&protected_programm[i][0],4);
          }
        }else{
          for(int i=0;i<16;i++){
             memcpy(c_programm[i],&unprotected_programm[i][0],4);
          }
       }
      }else{
        Serial.println(F("Unprotected code not found"));
        for(int i=0;i<13;i++){
             memcpy(c_programm[i],&protected_programm[i][0],4);
        }
      }
      reg_exec_flags |= EXECFLAG_PROGSET;
      Serial.println(F("Starting execution"));
    }
  }else{
    if(reg_slp){
      reg_slp--;
      return;
    }
    uint8_t  c_page = reg_eip / 4;
    uint8_t  c_word = reg_eip % 4;

    if(!reg_cop){
      op_t* op = find_op(c_programm[c_page][c_word]);
      if(!op){
        Serial.print(F("Failed to exec: Invalid operator 0x"));
        Serial.print(c_programm[c_page][c_word],HEX);
        Serial.println(F("!"));
        end_handlr(0);
        return;
      }
      if(op->flags & OPFLAG_NOARG){
        if(process_op(op)){
          Serial.println(F("Failed to exec: Operator failure!"));
          end_handlr(0);
          return;
        }
      }else{
        reg_cop = op;
      }
    }else{
      reg_carg = c_programm[c_page][c_word];
      if(process_op(reg_cop)){
        Serial.println(F("Failed to exec: Operator failure!"));
        end_handlr(0);
      }
      reg_cop = 0;
      reg_carg = 0;
    }

    reg_eip++;

    if(reg_eip >= 64){
      end_handlr(0);
    }
  }
  
}
